# MIT License
#
# Copyright (c) 2022-2025 Jérémie Chabloz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import subprocess
import os.path
import shutil
import socket
import logging


def find_free_port():
    """Find a free port on localhost

    The implementation for this function is not very elegant, but it does the
    job... It uses the property of :py:meth:`socket.socket.bind` method to bind
    the socket to a randomly-assigned free port when provided a :code:`('', 0)`
    address argument. Using :py:meth:`socket.socket.getsockname` is then used
    to retrieve the corresponding port number. Since the bound socket is closed
    within the function, it is assumed that the same port number should also be
    free again; this is where the weakness of this method lies, since race
    conditions cannot be fully excluded.

    Returns:
        int: A free port number
    """
    with socket.socket() as s:
        s.bind(('', 0))
        retval = s.getsockname()[1]
    return retval


def _format_path(cwd, path):
    if os.path.isabs(path):
        return path
    return os.path.abspath(os.path.join(cwd, path))


def setup_sim_run(elab_cmd, sim_cmd, capture_output=True,
                  capture_logfile=None):
    """Run simulation setup commands.

    This command is e.g. used by :py:meth:`setup_sim` with elaboration and
    simulation commands formatted according to the provided arguments.

    Args:
        elab_cmd (list): Elaboration command. It has to be provided as a list
            of command and arguments (see subprocess documentation). If None,
            this step is skipped.
        sim_cmd (list): Simulation command. It has to be provided as a list
            of command and arguments (see subprocess documentation). This
            command is mandatory and cannot be None.
        capture_output (bool): Defines if stdout and stderr output
            are "captured" (i.e. not visible).
        capture_logfile (int): Valid file descriptor to capture stdout and
            stderr

    Returns:
        subprocess.Popen: Process object of the launched simulation
    """

    if elab_cmd:
        subprocess.check_call(elab_cmd)

    if sim_cmd is None:
        raise ValueError("Simulation command and arguments is mandatory")

    if capture_output:
        pop = subprocess.Popen(
            sim_cmd,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL
        )
    else:
        if capture_logfile:
            pop = subprocess.Popen(
                sim_cmd, stdout=capture_logfile, stderr=capture_logfile)
        else:
            pop = subprocess.Popen(sim_cmd)

    logging.info(f"Launched simulation with PID {pop.pid}")
    return pop


def setup_sim(vpi_libpath, *src_files, cwd=".", vvp_filepath=None,
              vvp_logpath="vvp.log", ivl_exec=None, ivl_args=None,
              vvp_exec=None, vvp_args=None, vvp_postargs=None,
              capture_output=True, capture_logfile=None):
    """Set up Icarus simulation by elaborating the design with :code:`iverilog`
    and launching the simulation with :code:`vvp`. Uses
    :py:meth:`setup_sim_run` to run the concatenated commands and arguments.

    Args:
        vpi_libpath (str): Path to the compiled Verisocks VPI library.
        src_files (str): Paths to all (verilog) source files to use for the
            simulation. All files have to be added as separate arguments.
        cwd (str): Reference path to be used for all paths provided as relative
            paths.
        vvp_filepath (str): Path to the elaborated VVP file (iverilog output).
            If None (default), "sim.vvp" will be used.
        vvp_logpath (str): Path to a simulation logfile. Default="vvp.log". If
            None, no logfile shall be created.
        ivl_exec (str): Path to :code:`iverilog` executable (absolute path). If
            None (default), it is assumed to be defined in the system path.
        ivl_args (str, list(str)): Arguments to :code:`iverilog` executable.
        vvp_exec (str): Path to :code:`vvp` executable (absolute path). If None
            (default), it is assumed to be defined in the system path.
        vvp_args (list(str)): Arguments to :code:`vvp` executable.
        vvp_postargs (str, list(str)): (Post-)arguments to :code:`vvp`
            executable. In order to dump waveforms to an FST file, this should
            be configured as "-fst".
        capture_output (bool): Defines if stdout and stderr output
            are "captured" (i.e. not visible).
        capture_logfile (int): Valid file descriptor to capture stdout and
            stderr

    Returns:
        subprocess.Popen: Process object of the launched simulation
    """

    vpi_libpath = _format_path(cwd, vpi_libpath)
    if not os.path.isfile(vpi_libpath):
        raise FileNotFoundError(f"Could not find {vpi_libpath}")

    src_file_paths = []
    for src_file in src_files:
        src_file_path = _format_path(cwd, src_file)
        if not os.path.isfile(src_file_path):
            raise FileNotFoundError(f"File {src_file_path} not found")
        src_file_paths.append(src_file_path)

    if vvp_filepath:
        vvp_outfile = _format_path(cwd, vvp_filepath)
    else:
        vvp_outfile = _format_path(cwd, "sim.vvp")

    # Elaboration with iverilog
    if ivl_exec:
        ivl_cmd = [_format_path(cwd, ivl_exec)]
    else:
        ivl_cmd = [shutil.which("iverilog")]

    if ivl_args is None:
        ivl_args = []
    elif isinstance(ivl_args, str):
        ivl_args = [ivl_args]

    ivl_cmd += [
        "-o", vvp_outfile,
        "-Wall",
        *ivl_args,
        *src_file_paths
    ]

    # Simulation with vvp
    if vvp_exec:
        vvp_cmd = [_format_path(cwd, vvp_exec)]
    else:
        vvp_cmd = [shutil.which("vvp")]

    if vvp_logpath:
        vvp_cmd += ["-l" + _format_path(cwd, vvp_logpath)]
    if vvp_args is None:
        vvp_args = []
    elif isinstance(vvp_args, str):
        vvp_args = [vvp_args]
    if vvp_postargs is None:
        vvp_postargs = []
    elif isinstance(vvp_postargs, str):
        vvp_postargs = [vvp_postargs]

    vvp_cmd += [
        "-m", vpi_libpath,
        *vvp_args,
        vvp_outfile,
        *vvp_postargs
    ]

    pop = setup_sim_run(ivl_cmd, vvp_cmd, capture_output, capture_logfile)
    return pop



# MIT License
#
# Copyright (c) 2022-2025 Jérémie Chabloz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import argparse
import pathlib
import yaml
from os.path import isabs, abspath, dirname, join, relpath
from mako.template import Template


cwd = abspath(dirname(__file__))
tmpl_path = join(cwd, "templates")


def render_template(template_file, output_file, **kwargs):
    """Render Mako template

    Args:
        template_file (str): Path to template file
        output_file (str): Path to rendered output file
        kwargs: Keyword arguments required by the template
    """
    tmpl = Template(filename=template_file)
    tmpl_rendered = tmpl.render_unicode(
        template_filename=relpath(template_file, cwd), **kwargs)
    with open(output_file, 'w', encoding='utf-8') as f:
        f.write(tmpl_rendered)


def main():
    # Define arguments
    parser = argparse.ArgumentParser(description="""
This script generates automatically a Makefile, a C++ top testbench source file
and a Verilator configuration file to declare public variables for using
Verilator while integrating Verisocks. The files are generated based on Mako
templates and a YAML configuration file.
""")
    parser.add_argument('config', type=pathlib.Path,
                        help="YAML configuration file")
    parser.add_argument('--templates-dir', '-t', type=pathlib.Path,
                        default=tmpl_path,
                        help="Templates directory if alternative \
templates shall be used")
    parser.add_argument('--makefile', type=pathlib.Path, default="Makefile",
                        help="Rendered makefile name (default: Makefile)")
    parser.add_argument('--testbench-file', type=pathlib.Path,
                        default="test_main.cpp",
                        help="Rendered C++ testbench file (default: \
test_main.cpp)")
    parser.add_argument('--variables-file', type=pathlib.Path,
                        default="variables.vlt",
                        help="Rendered Verilator configuration file for \
public variables (default:variables.vlt)")
    parser.add_argument('--makefile-only', action='store_true',
                        help="Render makefile only (unless any other *-only \
option is being used)")
    parser.add_argument('--tb-only', action='store_true',
                        help="Render tesbench file only (unless any other \
*-only option is being used)")
    parser.add_argument('--vlt-only', action='store_true',
                        help="Render variables file only (unless any other \
*-only option is being used)")
    args = parser.parse_args()

    render_makefile = args.makefile_only or (
        not args.tb_only and not args.vlt_only)
    render_tb = args.tb_only or (
        not args.makefile_only and not args.vlt_only)
    render_vlt = args.vlt_only or (
        not args.makefile_only and not args.tb_only)

    # Format paths to templates relative to this file
    template_mk = join(args.templates_dir, "Makefile.mako")
    template_cpp = join(args.templates_dir, "test_main.cpp.mako")
    template_vlt = join(args.templates_dir, "variables.vlt.mako")

    # Load JSON configuration file
    with open(args.config, 'r') as f:
        cfg = yaml.safe_load(f)

    # Format relative paths in config file
    def format_path(x):
        if isabs(x):
            return x
        return abspath(join(dirname(args.config), x))
    for k in ["verisocks_root", "verilator_root", "verilator_path"]:
        format_path(cfg['config'][k])

    # Add C++ top testbench file at the front of C++ sources list
    if 'cpp_src_files' not in cfg['config']:
        cfg['config']['cpp_src_files'] = []
    cfg['config']['cpp_src_files'] = (
        [str(args.testbench_file)] + cfg['config']['cpp_src_files'])

    # Add verilator configuration file for public variables at the front of the
    # Verilog sources list
    if 'variables' in cfg:
        cfg['config']['verilog_src_files'] = (
            [str(args.variables_file)] + cfg['config']['verilog_src_files'])

    if render_makefile:
        if 'variables' in cfg:
            vlt_file=str(args.variables_file)
        else:
            vlt_file=None
        render_template(template_mk, args.makefile,
                        target_file=str(args.makefile),
                        config_file=str(args.config),
                        tb_file=str(args.testbench_file),
                        vlt_file=vlt_file,
                        **cfg['config'])
    if render_tb:
        if 'variables' in cfg:
            render_template(template_cpp, args.testbench_file,
                            **cfg['config'], variables=cfg['variables'])
        else:
            render_template(template_cpp, args.testbench_file,
                            **cfg['config'], variables=None)
    if render_vlt and ('variables' in cfg):
        render_template(template_vlt, args.variables_file,
                        variables=cfg['variables'])


if __name__ == "__main__":
    main()



from verisocks.verisocks import Verisocks, VerisocksError
from verisocks.utils import setup_sim, find_free_port
import os.path
import pytest
import logging
import re
import socket

# Note
# To run with coverage, use
# coverage run --source=../verisocks --branch -m pytest
# coverage report or coverage html


# Parameters
HOST = socket.gethostbyname("localhost")
LIBVPI = "../../build/verisocks.vpi"  # Relative path to this file!
VS_TIMEOUT = 2


# Expectations
sim_info_product = "Icarus Verilog"
sim_info_version = r"1[0-3]\.[0-9]+"


def setup_test(port):
    pop = setup_sim(
        LIBVPI,
        "test_0.v",
        cwd=os.path.dirname(__file__),
        ivl_args=[
            f"-DNUM_PORT={port}",
            f"-DVS_TIMEOUT={VS_TIMEOUT}"
        ]
    )
    return pop


@pytest.fixture(params=[True, False])
def vs(request):
    # Setup
    port = find_free_port()
    pop = setup_test(port)
    _vs = Verisocks(HOST, port, use_uuid=request.param)
    _vs.connect()
    yield _vs
    # Teardown
    try:
        _vs.finish()
    except ConnectionError:
        logging.warning("Connection error - Cannot send finish command")
    _vs.close()
    pop.communicate(timeout=10)


def test_connect(vs):
    """Tests simple connection to Icarus simulator with Verisocks server"""
    assert vs._connected


def test_already_connected(vs, caplog):
    caplog.set_level(logging.INFO)
    caplog.clear()
    vs.connect()
    log_record = caplog.record_tuples[0]
    assert log_record[1] == logging.INFO
    assert log_record[2] == "Socket already connected"


def test_connect_error():
    """Tests trying to connect to a non-running server"""
    port = find_free_port()
    with pytest.raises(ConnectionError):
        vs = Verisocks(HOST, port)
        vs.connect(trials=1)
    vs.close()


def test_info(vs):
    """Tests the info command"""
    answer = vs.info("This is a test")
    assert answer["type"] == "ack"


def test_get_sim_info(vs):
    answer = vs.get("sim_info")
    assert answer["type"] == "result"
    assert answer["product"] == sim_info_product
    assert re.match(sim_info_version, answer["version"])


def test_get_sim_time(vs):
    """Tests Verisocks get(sel=sim_time) function"""
    answer = vs.get("sim_time")
    assert answer["type"] == "result"
    assert answer["time"] == 0.0

    answer = vs.run("until_time", time=101.3, time_unit="us")
    assert answer["type"] == "ack"

    answer = vs.get("sim_time")
    assert answer["type"] == "result"
    assert answer["time"] == 101.3e-6


def test_get_value(vs):
    """Tests Verisocks get(sel="value") function"""

    # Get an integer parameter value
    answer = vs.get("value", path="main.int_param")
    assert answer["type"] == "result"
    assert answer["value"] == 598402

    # Get a real parameter value
    answer = vs.get(sel="value", path="main.fclk")
    assert answer["type"] == "result"
    assert answer["value"] == 1.01

    # Get a reg value
    answer = vs.run("for_time", time=100, time_unit="us")
    assert answer["type"] == "ack"
    answer = vs.get(sel="value", path="main.clk")
    assert answer["type"] == "result"
    assert answer["value"] == 1

    # Get a reg[] value
    answer = vs.get(sel="value", path="main.count")
    assert answer["type"] == "result"
    assert answer["value"] == 101

    # Get a memory array value
    answer = vs.get(sel="value", path="main.count_memory")
    assert answer["type"] == "result"
    assert answer["value"] == [
        96, 97, 98, 99, 100, 85, 86, 87,
        88, 89, 90, 91, 92, 93, 94, 95]

    answer = vs.get(sel="value", path="main.count_memory[3]")
    assert answer["type"] == "result"
    assert answer["value"] == 99

    # Error case: wrong path
    with pytest.raises(VerisocksError):
        answer = vs.get(sel="value", path="wrong_path")
        assert answer["type"] == "error"


def test_get_type(vs):
    """Tests Verisocks get(sel="type") function"""

    # Get an integer parameter type
    answer = vs.get("type", path="main.int_param")
    assert answer["type"] == "result"
    assert answer["vpi_type"] == 41  # vpiParameter (see vpi_user.h)

    # Get a real parameter type
    answer = vs.get(sel="type", path="main.fclk")
    assert answer["type"] == "result"
    assert answer["vpi_type"] == 41  # vpiParameter (see vpi_user.h)

    # Get a reg type
    answer = vs.get(sel="type", path="main.clk")
    assert answer["type"] == "result"
    assert answer["vpi_type"] == 48  # vpiReg (see vpi_user.h)

    # Get a reg[] type
    answer = vs.get(sel="type", path="main.count")
    assert answer["type"] == "result"
    assert answer["vpi_type"] == 48  # vpiReg (see vpi_user.h)

    # Get a memory array type
    answer = vs.get(sel="type", path="main.count_memory")
    assert answer["type"] == "result"
    assert answer["vpi_type"] == 29  # vpiMemory (see vpi_user.h)

    # Error case: wrong path
    with pytest.raises(VerisocksError):
        answer = vs.get(sel="type", path="wrong_path")
        assert answer["type"] == "error"


def test_run_for_time(vs):
    """Tests Verisocks run(cb="for_time") function"""

    # Get initial time
    answer = vs.get(sel="sim_time")
    assert answer["type"] == "result"
    prev_sim_time = answer["time"]

    # For time in ps
    answer = vs.run(cb="for_time", time=6, time_unit="ps")
    assert answer["type"] == "ack"
    answer = vs.get(sel="sim_time")
    assert answer["type"] == "result"
    assert answer["time"] - prev_sim_time == pytest.approx(6e-12)
    prev_sim_time = answer["time"]

    # For time in ns
    answer = vs.run(cb="for_time", time=7, time_unit="ns")
    assert answer["type"] == "ack"
    answer = vs.get(sel="sim_time")
    assert answer["type"] == "result"
    assert answer["time"] - prev_sim_time == pytest.approx(7e-9)
    prev_sim_time = answer["time"]

    # For time in us
    answer = vs.run(cb="for_time", time=8, time_unit="us")
    assert answer["type"] == "ack"
    answer = vs.get(sel="sim_time")
    assert answer["type"] == "result"
    assert answer["time"] - prev_sim_time == pytest.approx(8e-6)
    prev_sim_time = answer["time"]

    # For time in ms
    answer = vs.run(cb="for_time", time=0.23, time_unit="ms")
    assert answer["type"] == "ack"
    answer = vs.get(sel="sim_time")
    assert answer["type"] == "result"
    assert answer["time"] - prev_sim_time == pytest.approx(0.23e-3)
    prev_sim_time = answer["time"]

    # For time in s
    answer = vs.run(cb="for_time", time=0.00017, time_unit="s")
    assert answer["type"] == "ack"
    answer = vs.get(sel="sim_time")
    assert answer["type"] == "result"
    assert answer["time"] - prev_sim_time == pytest.approx(0.17e-3)
    prev_sim_time = answer["time"]


def test_run_to_next(vs):
    """Tests Verisocks run(cb="to_next") function"""

    # To next
    answer = vs.run(cb="to_next")
    assert answer["type"] == "ack"
    answer = vs.get(sel="sim_time")
    assert answer["type"] == "result"
    assert answer["time"] == pytest.approx(0.1e-6)


def test_run_until_time(vs):
    """Tests Verisocks run(cb="until_time") function"""

    # Until time in ps
    answer = vs.run(cb="until_time", time=174, time_unit="ps")
    assert answer["type"] == "ack"
    answer = vs.get(sel="sim_time")
    assert answer["type"] == "result"
    assert answer["time"] == pytest.approx(174e-12)

    # Until time in ns
    answer = vs.run(cb="until_time", time=215.4, time_unit="ns")
    assert answer["type"] == "ack"
    answer = vs.get(sel="sim_time")
    assert answer["type"] == "result"
    assert answer["time"] == pytest.approx(215.4e-9)

    # Until time in us
    answer = vs.run(cb="until_time", time=1.593, time_unit="us")
    assert answer["type"] == "ack"
    answer = vs.get(sel="sim_time")
    assert answer["type"] == "result"
    assert answer["time"] == pytest.approx(1.593e-6)

    # Until time in ms
    answer = vs.run(cb="until_time", time=0.456, time_unit="ms")
    assert answer["type"] == "ack"
    answer = vs.get(sel="sim_time")
    assert answer["type"] == "result"
    assert answer["time"] == pytest.approx(0.456e-3)

    # Until time in s
    answer = vs.run(cb="until_time", time=0.6e-3, time_unit="s")
    assert answer["type"] == "ack"
    answer = vs.get(sel="sim_time")
    assert answer["type"] == "result"
    assert answer["time"] == pytest.approx(0.6e-3)

    # Error: until time in the past
    with pytest.raises(VerisocksError):
        answer = vs.run(cb="until_time", time=0.4e-3, time_unit="s")
        assert answer["type"] == "error"


def test_run_until_change(vs):
    """Tests Verisocks run(cb="until_change") function"""

    answer = vs.run(cb="until_change", path="main.count", value=137)
    assert answer["type"] == "ack"
    answer = vs.get(sel="value", path="main.count")
    assert answer["type"] == "result"
    assert answer["value"] == 137

    answer = vs.run(cb="until_change", path="main.counter_end")
    assert answer["type"] == "ack"
    answer = vs.get(sel="value", path="main.count")
    assert answer["type"] == "result"
    assert answer["value"] == 255


def test_set(vs):
    """Tests Verisocks set() function"""
    # Set a reg
    answer = vs.run(cb="until_time", time=10, time_unit="us")
    assert answer["type"] == "ack"
    answer = vs.set(path="main.count", value=125)
    assert answer["type"] == "ack"
    answer = vs.get(sel="value", path="main.count")
    assert answer["type"] == "result"
    assert answer["value"] == 125

    # Set a memory array
    answer = vs.set(path="main.count_memory", value=list(range(16)))
    assert answer["type"] == "ack"
    answer = vs.get(sel="value", path="main.count_memory")
    assert answer["type"] == "result"
    assert answer["value"] == list(range(16))

    answer = vs.set(path="main.count_memory[6]", value=37)
    assert answer["type"] == "ack"
    answer = vs.get(sel="value", path="main.count_memory[6]")
    assert answer["type"] == "result"
    assert answer["value"] == 37

    # Set an event
    answer = vs.set(path="main.counter_end")
    assert answer["type"] == "ack"


def test_read_not_expected(vs):
    """Tests what happens when a read function is requested while there are no
    messages expected.
    """
    assert vs.read() is False


def test_sim_finishes(vs):
    """Tests the case where the Icarus simulation runs out before the specified
    callback event occurs
    """
    with pytest.raises(VerisocksError):
        _ = vs.run(cb="until_time", time=1200, time_unit="us")


def test_exit():
    port = find_free_port()
    setup_test(port)
    with Verisocks(HOST, port) as vs:
        answer = vs.exit()
        assert answer["type"] == "ack"


def test_context_manager():
    port = find_free_port()
    pop = setup_test(port)
    with Verisocks(HOST, port) as vs:
        assert vs._connected
        answer = vs.finish()
        assert answer["type"] == "ack"

    pop.wait(timeout=10)



# MIT License
#
# Copyright (c) 2022-2025 Jérémie Chabloz
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import socket
import logging
import struct
import json
from enum import Enum, auto
from time import sleep
from uuid import UUID, uuid4


class VsRxState(Enum):
    """Enumerated states for the RX state machine

    Possible values are:

        - ``RX_INIT``: Starting state
        - ``RX_PRE_HDR``: RX message pre-header has been scanned
        - ``RX_HDR``: RX message header has been scanned
        - ``RX_DONE``: RX message content has been scanned
        - ``ERROR``: Error state
    """
    RX_INIT = auto()     # Starting state
    RX_PRE_HDR = auto()  # RX message pre-header scanned
    RX_HDR = auto()      # RX message header scanned
    RX_DONE = auto()     # RX message content scanned
    ERROR = auto()       # Error state


class VerisocksError(Exception):
    """Base class for exceptions in Verisocks"""
    pass


class Verisocks:
    """Verisocks client class.

    Args:
        host (str): Server host IP address, default="127.0.0.1"
        port (int): Server port number, default=5100
        timeout (float): Socket timeout (base value),
                            in seconds (default=120)
        connect_trials (int): Number of consecutive connections to be attempted
            when the method
            :py:meth:`connect()<verisocks.verisocks.Verisocks.connect>` is
            being used. This value can be overriden by the method's own
            ``trials`` argument.
        connect_delay (float): Delay to be applied before attempting a new
            connection trial when the method
            :py:meth:`connect()<verisocks.verisocks.Verisocks.connect>` is
            being used. This value can be overriden by the method's own
            ``delay`` argument.
        use_uuid (bool): Use transactions UUID. If true (default), an UUID
            number will be added to the request header and will be verified to
            match when the corresponding answer is received.

    Note:
        For certain methods, a specific timeout value can be passed as
        argument. If a value (other than ``None``) is provided, the socket
        timeout will be modified only temprorarily and then restored to the
        timeout value provided as argument at the class instantiation.
    """

    PRE_HDR_LEN = 2  # Pre-header length in bytes
    READ_BUFFER_LEN = 4096

    def __init__(self, host="127.0.0.1", port=5100, timeout=120.0,
                 connect_trials=10, connect_delay=0.05, use_uuid=True):
        """Verisocks class constructor
        """
        # Connection address and status
        self._connected = False
        self.address = (host, port)
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.setblocking(True)
        self._timeout = None
        if timeout:
            self._timeout = timeout
            self.sock.settimeout(timeout)
        self.connect_trials = connect_trials
        self.connect_delay = connect_delay
        self.use_uuid = use_uuid
        self.uuid = None

        # RX variables
        self._rx_buffer = b""
        self._rx_header_len = None
        self._rx_state = VsRxState.RX_INIT
        self._rx_expected = 0
        self.rx_header = None
        self.rx_content = None

        # TX variables
        self._tx_buffer = b""
        self._tx_msg_len = []

    def connect(self, trials=None, delay=None):
        """Connect to server socket.

        If the client is already connected to a server socket, nothing happens.
        Otherwise, the client attempts to connect to the server as defined by
        the address and port provided to the class constructor. The method will
        apply a delay prior each connection trial and will retry a number of
        times if unsuccessful.

        Args:
            trials (int): Maximum number of connection trials to be attempted.
                If None, the value of the ``connect_trials`` argument passed to
                the constructor is being used.
            delay (float): Delay to be applied prior each connection trial. If
                None, the value of the ``connect_delay`` argument passed to the
                constructor is being used.

        Raises:
            ConnectionError: All the successive connection trials have been
                unsucessful
        """

        if trials is None:
            trials = self.connect_trials
        if delay is None:
            delay = self.connect_delay

        if not self._connected:

            if self.sock is None:
                self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                self.sock.setblocking(True)
                self.sock.settimeout(self._timeout)

            logging.info(f"Attempting connection to {self.address}")
            trial = 0
            while trial < trials:
                try:
                    self.sock.connect(self.address)
                    logging.info(f"Socket connected after {trial + 1} trials")
                    self._connected = True
                    break
                except ConnectionError:
                    sleep(delay)
                    trial += 1
            if trial >= trials:
                raise ConnectionError(
                    f"Connection unsucessful after {trial} trials")
        else:
            logging.info("Socket already connected")

    def _read(self, timeout=None):
        """Reads from socket to RX buffer (private).

        Args:
            timeout (float): Timeout in seconds (default=None). If None, the
            base value as defined within the class constructor applies.

        Raises:
            ConnectionError: no data is received (most likely the socket is
                closed).
        """
        if timeout:
            self.sock.settimeout(timeout)
        data = self.sock.recv(self.READ_BUFFER_LEN)
        if data:
            self._rx_buffer += data
        else:
            raise ConnectionError
        if timeout:
            self.sock.settimeout(self._timeout)

    def _write(self, num_bytes, num_trials=10):
        """Write TX buffer to socket (private method)

        Args:
            num_bytes (int): Number of bytes to send

        Raises:
            ConnectionError: no data is written (most likely the socket is
                closed).
        """
        sent = 0
        trials = 0
        while ((sent < num_bytes) and (trials < num_trials)):
            sent += self.sock.send(self._tx_buffer[sent:num_bytes])
            if (sent == 0):
                self._connected = False
                raise ConnectionError
            trials += 1

        if (sent == num_bytes):
            self._tx_buffer = self._tx_buffer[sent:]
            logging.debug(
                f"Sent {num_bytes} bytes on socket in {trials} trial(s)")
        else:
            logging.error("Did not succeed to write message to socket")
            self.close()

    def _json_encode(self, obj, encoding="utf-8"):
        """Encode a JSON object as a bytes object (private method)

        Args:
            obj (json): JSON object
            encoding (str): Encoding to be used, default=utf-8

        Returns:
            bytes: Dumped and encoded content
        """
        return json.dumps(obj, ensure_ascii=False).encode(encoding)

    def _read_pre_header(self):
        """Parse RX buffer for pre-header value (private method).
        """
        if (self._rx_state is not VsRxState.RX_INIT):
            raise ValueError("ERROR: Inconsistent state")
        if not (len(self._rx_buffer) >= self.PRE_HDR_LEN):
            return
        self._rx_header_len = struct.unpack(
            ">H", self._rx_buffer[:self.PRE_HDR_LEN]
        )[0]
        self._rx_buffer = self._rx_buffer[self.PRE_HDR_LEN:]
        self._rx_state = VsRxState.RX_PRE_HDR

    def _read_header(self):
        """Parse RX buffer for header (private method).
        """
        if (self._rx_state is not VsRxState.RX_PRE_HDR):
            raise RuntimeError("ERROR: Inconsistent state")
        header_len = self._rx_header_len
        if not (len(self._rx_buffer) >= header_len):
            return
        logging.debug("Received message header: " +
                      self._rx_buffer[:header_len].decode("utf-8"))
        self.rx_header = json.loads(
            self._rx_buffer[:header_len].decode("utf-8"))
        self._rx_buffer = self._rx_buffer[header_len:]
        header_keys = [
            "content-length",
            "content-type",
            "content-encoding"
        ]
        if self.use_uuid:
            header_keys.append("uuid")
        for k in header_keys:
            if k not in self.rx_header:
                raise ValueError(f"Missing required header field '{k}'.")
        self._rx_state = VsRxState.RX_HDR

        if self.use_uuid:
            self.rx_uuid = UUID(self.rx_header['uuid'])
            if (self.rx_uuid != self.uuid):
                raise VerisocksError("Inconsistent transaction UUID values")
        else:
            if "uuid" in self.rx_header:
                raise VerisocksError("Unexpected transaction UUID")

    def _read_content(self):
        """Parse RX buffer for content (private method).
        """
        if (self._rx_state is not VsRxState.RX_HDR):
            raise RuntimeError("ERROR: Inconsistent state")
        content_len = self.rx_header["content-length"]
        if not (len(self._rx_buffer) >= content_len):
            return
        data = self._rx_buffer[:content_len]
        self._rx_buffer = self._rx_buffer[content_len:]
        logging.debug("Received message content: " + repr(data))

        # Process content depending on type declared in header
        if (self.rx_header["content-type"] == "text/plain"):
            encoding = self.rx_header["content-encoding"]
            self.rx_content = data.decode(encoding)
        elif (self.rx_header["content-type"] == "application/json"):
            encoding = self.rx_header["content-encoding"]
            self.rx_content = json.loads(data.decode(encoding))
        elif (self.rx_header["content-type"] == "application/octet-stream"):
            self.rx_content = data
        else:
            raise ValueError("Value for 'content-type' not recognized")
        self._rx_state = VsRxState.RX_DONE

    def _queue_message(self, request):
        """Create message and add it to the TX queue buffer.

        Args:
            request (dict): Dictionary with the following keys: ``"content"``,
                ``"type"`` and ``"encoding"`` (unless ``"type"`` is
                ``"application/octet-stream"``, in which case the encoding is
                not applicable).
        """
        content = request["content"]
        content_type = request["type"]

        # Prepare content
        if (content_type == "text/plain"):
            content_encoding = request["encoding"]
            content_bytes = bytes(content, encoding=content_encoding)
        elif (content_type == "application/json"):
            content_encoding = request["encoding"]
            content_bytes = self._json_encode(content, content_encoding)
        elif (content_type == "application/octet-stream"):
            content_bytes = bytes(content)
        else:
            raise ValueError(
                f"Value for 'content_type' {content_type} not recognized")

        # Create header
        if content_type == "application/octet-stream":
            json_header = {
                "content-type": content_type,
                "content-length": len(content_bytes)
            }
        else:
            json_header = {
                "content-type": content_type,
                "content-encoding": content_encoding,
                "content-length": len(content_bytes)
            }

        # Transaction UUID
        if self.use_uuid:
            self.uuid = uuid4()
            json_header['uuid'] = self.uuid.urn.split(":")[-1]
        message_header = self._json_encode(json_header, "utf-8")

        # Adjust pre-header
        message_pre_header = struct.pack(">H", len(message_header))

        # Compose full message and add it to the buffer
        message = message_pre_header + message_header + content_bytes
        self._tx_buffer += message
        self._tx_msg_len.append(len(message))
        logging.debug(f"Queuing message header (length {len(message_header)})\
: {repr(message_header)}")
        logging.debug(f"Queuing message content: {repr(content_bytes)}")

    def read(self, num_trials=10, timeout=None):
        """Proceed to read and scan returned TCP messages

        Args:
            num_trials (int): Maximum number of trials. Default=10.
            timeout (float): Timeout in seconds (default=None). If None, the
                base value as defined within the class constructor applies.

        Returns:
            bool: Status. ``True`` if successful, ``False`` if error.
        """
        if self._rx_expected:
            if not self._connected:
                self.connect()
            if not self._rx_buffer:
                self._read(timeout)
            trials = 0
            while (trials < num_trials):
                if (self._rx_state is VsRxState.RX_INIT):
                    self._read_pre_header()
                if (self._rx_state is VsRxState.RX_PRE_HDR):
                    self._read_header()
                if (self._rx_state is VsRxState.RX_HDR):
                    self._read_content()
                if (self._rx_state is VsRxState.RX_DONE):
                    self._rx_expected -= 1
                    self._rx_state = VsRxState.RX_INIT
                    logging.debug(f"Read procedure successful. \
Still {self._rx_expected} messages expected.")
                    return True
                self._read(timeout)
                trials += 1
            self._rx_state = VsRxState.ERROR
            logging.error("Read procedure unsuccessful")
            return False
        else:
            logging.warning("No expected message. Cancelling read procedure.")
            return False

    def write(self, all=True):
        """Writes/sends the current content of the TX buffer to the socket.

        Args:
            all (bool): If True (default), sends all queued message. Otherwise,
                it only sends the oldest queued message in the buffer.
        """
        if self._tx_buffer:  # Check that the TX buffer is not empty
            if not self._connected:
                self.connect()
            if all:
                self._write(len(self._tx_buffer))
                self._rx_expected += len(self._tx_msg_len)
                self._tx_msg_len.clear()
            else:
                self._write(self._tx_msg_len.pop(0))
                self._rx_expected += 1
        else:
            logging.warning("TX buffer is empty. No message to transmit.")

    def send(self, **cmd):
        """Sends a message with a JSON content.

        This command will accept and use any command content defined as keyword
        arguments (e.g. ``cmd="get"``, ``sel="sim_info"``). This content will
        directly be used as the sent TCP command content.

        Keyword Arguments:
            timeout (float): Socket timeout configuration value in seconds.
                If None (default), the class instance default value is used.

        Returns:
            JSON object: Content of returned message.
        """

        if "timeout" in cmd:
            timeout = cmd.pop("timeout")
        else:
            timeout = None

        self._queue_message({
            "type": "application/json",
            "encoding": "utf-8",
            "content": cmd
        })
        self.write()

        if (self.read(10, timeout)):
            if self.rx_content["type"] == "error":
                raise VerisocksError(self.rx_content["value"])
            return self.rx_content
        else:
            return None

    def send_cmd(self, command, **kwargs):
        """Sends a command. Equivalent to :code:`send(command=command, ...)`.

        Args:
            command (str): Command name (e.g. ``"get"``)
            **kwargs: Command keyword arguments (e.g. ``sel="sim_info"``)

        Keyword Arguments:
            timeout (float): Socket timeout configuration value in seconds.
                If None (default), the class instance default value is used.

        Returns:
            JSON object: Content of returned message
        """
        return self.send(command=command, **kwargs)

    def run(self, cb, **kwargs):
        """Sends a :keyword:`run <sec_tcp_cmd_run>` command request to the
        Verisocks server.

        Equivalent to :code:`send_cmd("run", cb=cb, ...)`. This command gives
        the focus back to the simulator and lets it run until the specified
        callback condition is met (see arguments).

        Args:
            cb (str): Callback type.

        Keyword Arguments:
            timeout (float): Socket timeout configuration value in seconds.
                If None (default), the class instance default value is used.

        Returns:
            JSON object: Content of returned message

        Note:
            The parameter `cb` can be either:

            * ``"for_time"``: run for a given amount of time
            * ``"until_time"``: run until a specified time
            * ``"until_change"``: run until a specific value changes
            * ``"to_next"``: run until the beginning of the next time step

            If `cb` is ``"for_time"`` or ``"until_time"``, the following
            keyword arguments are further expected:

            * **time** (float): Time value
            * **time_unit** (str): Time unit (s, ms, us, ns, ps or fs). Be
              aware that depending on the simulator time resolution, the
              provided time value may be truncated.

            If `cb` is ``"until_change"``, the following keyword argument(s)
            are further expected:

            * **path** (str): Path to verilog object used for the callback
            * **value** (number): Condition on the verilog object's
              value for the callback to be executed. This argument is not
              required if the path corresponds to a named event.

            If `cb` is ``"to_next"``, no further keyword argument is required.
        """

        return self.send(command="run", cb=cb, **kwargs)

    def set(self, path, **kwargs):
        """Sends a :keyword:`set <sec_tcp_cmd_set>` command request to the
        Verisocks server.

        Equivalent to :code:`send_cmd("set", path=path, **kwargs)`. This
        commands sets the value of a verilog object as defined by its path.

        Args:
            path (str): Path to the verilog object.

        Keyword Arguments:
            value (int, list): Value to be set. If the path corresponds to a
                verilog named event, this argument is not required. If the path
                corresponds to a verilog memory array, this argument needs to
                be provided as a list of the same length.
            timeout (float): Socket timeout configuration value in seconds.
                If None (default), the class instance default value is used.

        Returns:
            JSON object: Content of returned message
        """
        return self.send(command="set", path=path, **kwargs)

    def info(self, value):
        """Sends an :keyword:`info <sec_tcp_cmd_info>` command to the Verisocks
        server.

        This is a shortcut function, which is equivalent to
        :code:`send_cmd("info", ...)`. This command is used to send any text to
        the Verisocks server, which will then be streamed out to the VPI
        standard output.

        Args:
            value (str): Text to be sent to the VPI stdout

        Returns:
            JSON object: Content of returned message
        """
        return self.send(command="info", value=value)

    def get(self, sel, path=None):
        """Sends a :keyword:`get <sec_tcp_cmd_get>` command request to the
        Verisocks server.

        Equivalent to :code:`send_cmd("get", ...)`. This commands can be
        used to obtain different pieces of information from the Verisocks
        server.

        Args:
            sel (str): Selects which is the returned information.
            path (str): If `sel` is ``"value"`` or ``"type"``, path to the
                desired verilog object for which the value or type is to be
                returned.

        Note:
            The argument `sel` can take the following values:

            * ``"sim_info"``: Gets the simulator information,
              returned with the keywords :code:`"product"` and
              :code:`"version"`.
            * ``"sim_time"``: Gets the simulator current time, returned
              with the keywords :code:`"time"`, in seconds.
            * ``"value"``: Gets the value for a verilog object.
            * ``"type"``: Gets the VPI type value for a verilog object.

        Returns:
            JSON object: Content of returned message
        """
        if path:
            return self.send(command="get", sel=sel, path=path)
        return self.send(command="get", sel=sel)

    def finish(self, timeout=None):
        """Sends a :keyword:`finish <sec_tcp_cmd_finish>` command to the
        Verisocks server that terminates the simulation (and therefore also
        closes the Verisocks server itself). The connection is closed as well
        by the function as a clean-up.
        """
        retval = self.send(command="finish", timeout=timeout)
        self.close()
        return retval

    def stop(self, timeout=None):
        """Sends a :keyword:`stop <sec_tcp_cmd_stop>` command to the Verisocks
        server.

        The ``"stop"`` command stops the simulation. The Verisocks server
        socket is not closed, but the simulation has to be restarted for any
        new request to be processed.

        Args:
            timeout (int, None): Timeout in seconds to apply for the execution
                of the command.
        """
        return self.send(command="stop", timeout=timeout)

    def exit(self):
        """Sends an :keyword:`exit <sec_tcp_cmd_exit>` command to the Verisocks
        server that gives back control to the simulator and closes the
        Verisocks server socket. The simulation runs to its end without having
        the possibility to take the control back from the simulator anymore.
        The connection is closed as well by the function."""
        retval = self.send(command="exit")
        self.close()
        return retval

    def close(self):
        """Close socket connection if still open."""
        if self._connected:
            logging.info("Closing socket connection")
            self.sock.close()
            self.sock = None
            self._connected = False

    def flush(self):
        """Flush RX and TX buffers."""
        logging.debug("Flushing RX and TX buffers")
        self._rx_buffer = b""
        self._tx_buffer = b""
        # self._rx_expected = 0
        self._tx_msg_len = []

    def __enter__(self):
        """Context manager - Entry function"""
        self.connect()
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        """Context manager - Exit function"""
        self.close()
        return False

# EOF
